import { system, world, ItemStack, BlockTypes } from "@minecraft/server" //scripts> npm i @minecraft/server@beta
import { ActionFormData, ModalFormData } from "@minecraft/server-ui"
import "./worldTicksEvents"

system.beforeEvents.watchdogTerminate.subscribe(s => { s.cancel = true })

function setPermutations(x, y, z, blockName, properties, values) {
    const props = { ...properties }
    for (const [key, val] of Object.entries(values)) {
        if (props.hasOwnProperty(key)) {
            props[key] = val
        }
    }
    const propStrings = Object.entries(props).map(([key, val]) => `"${key}": ${val}`);
    return `setblock ${x} ${y} ${z} ${blockName} [ ${propStrings.join(", ")} ]`;
}

world.afterEvents.itemUse.subscribe(({ itemStack: item, source }) => {
    if (item.typeId === 'effect99:item_hoover' && item.getLore()[0]) {
        const lore = item.getLore()

        if (Number(item.getLore()[0].match(/[\d.]+/g)[0])) {
            if (lore.includes('§r§l§cOff')) lore.splice(1, 1, '§r§l§aOn')
            else lore.splice(1, 1, '§r§l§cOff')
        }
        item.setLore(lore)

        for (const player of world.getAllPlayers()) {
            if (player.id === source.id) {
                player.getComponent('inventory').container.setItem(player.selectedSlot, item)
            }
        }
    }
})

world.afterEvents.playerPlaceBlock.subscribe(({ block, block: { x, y, z }, dimension }) => {
    if (block.typeId === 'effect99:storage_core_placer') {
        let valid = 1;
        for (let i = -2; i <= 2; i++) {
            for (let j = 0; j <= 2; j++) {
                for (let k = -2; k <= 2; k++) {
                    const block = dimension.getBlock({ x: x + i, y: y + j, z: z + k });
                    ['minecraft:air', 'effect99:storage_core_placer'].some(e => block.typeId.includes(e)) ? 0 : valid = 0
                }
            }
        }
        if (valid) {
            dimension.runCommand(`structure load storage_core ${x - 2} ${y} ${z - 2}`)
            dimension.spawnEntity('entity:storage_core', { x: x + 0.5, y: y + 2.5, z: z + 0.5 })
        }
        else {
            dimension.runCommand(`setblock ${x} ${y} ${z} air 0 destroy`)
            dimension.spawnEntity('entity:storage_core_zone', { x: x + 0.5, y: y, z: z + 0.5 })
        }
    }
})

system.afterEvents.scriptEventReceive.subscribe(s => {
    try {
        const block = s.sourceBlock
        const { x, y, z } = block
        const dimension = world.getDimension(s.sourceBlock.dimension.id)
        if (s.id == 'drill:block') {
            switch (Number(s.message)) {
                case 0: blockDestroy(x, y + 1, z, dimension.id, x, y, z); break
                case 1: blockDestroy(x, y - 1, z, dimension.id, x, y, z); break
                case 2: blockDestroy(x, y, z + 1, dimension.id, x, y, z); break
                case 3: blockDestroy(x, y, z - 1, dimension.id, x, y, z); break
                case 4: blockDestroy(x + 1, y, z, dimension.id, x, y, z); break
                case 5: blockDestroy(x - 1, y, z, dimension.id, x, y, z); break
            }
        }
        else if (s.id === 'is:lava') {
            switch (Number(s.message)) {
                case 1: if (dimension.getBlock({ x: x, y: y - 1, z: z - 1 }).typeId === 'minecraft:lava') isLava(); break
                case 2: if (dimension.getBlock({ x: x, y: y - 1, z: z + 1 }).typeId === 'minecraft:lava') isLava(); break
                case 3: if (dimension.getBlock({ x: x - 1, y: y - 1, z: z }).typeId === 'minecraft:lava') isLava(); break
                case 4: if (dimension.getBlock({ x: x + 1, y: y - 1, z: z }).typeId === 'minecraft:lava') isLava(); break
            }
            function isLava() {
                const permutation = block.permutation
                dimension.runCommand(setPermutations(x, y, z, block.typeId, permutation.getAllStates(), { "liquid:pumping": 2 }))
            }
        }
        else if (s.id === 'industrial:fan') {
            let vector, k, impulse, velocity

            for (let i = 1; i <= 10; i++) {
                const rotation = Number(s.message)
                switch (rotation) {
                    case 0: vector = { x: x, y: y + i, z: z }; k = { x: 0, z: 0, hS: 0, vS: 2.8 / (i + Math.random()) }; impulse = { x: 0, y: (Math.random() - 0.25) / i, z: 0 }; break
                    case 1: vector = { x: x, y: y - i, z: z }; k = { x: 0, z: 0, hS: 0, vS: -2.8 / (i + Math.random()) }; impulse = { x: 0, y: -1 / i, z: 0 }; break
                    case 2: vector = { x: x, y: y, z: z + i }; k = { x: 0, z: -1, hS: -2 / i, vS: 0 }; impulse = { x: 0, y: 0, z: 0.8 / i }; break
                    case 3: vector = { x: x, y: y, z: z - i }; k = { x: 0, z: 1, hS: -2 / i, vS: 0 }; impulse = { x: 0, y: 0, z: -0.8 / i }; break
                    case 4: vector = { x: x + i, y: y, z: z }; k = { x: -1, z: 0, hS: -2 / i, vS: 0 }; impulse = { x: 0.8 / i, y: 0, z: 0 }; break
                    case 5: vector = { x: x - i, y: y, z: z }; k = { x: 1, z: 0, hS: -2 / i, vS: 0 }; impulse = { x: -0.8 / i, y: 0, z: 0 }; break
                }
                if (dimension.getBlock(vector).isSolid || dimension.getBlock(vector).isLiquid) break
                for (const entity of block.dimension.getEntitiesAtBlockLocation(vector)) {
                    switch (rotation) {
                        case 0: ; case 1: velocity = entity.getVelocity().y <= -Math.random() + 0.5; break
                        case 2: velocity = entity.getVelocity().z >= -Math.random() + 0.5; break
                        case 3: velocity = entity.getVelocity().z <= -Math.random() + 0.5; break
                        case 4: velocity = entity.getVelocity().x >= -Math.random() + 0.5; break
                        case 5: velocity = entity.getVelocity().x <= -Math.random() + 0.5; break
                    }
                    const isPlayer = entity.typeId === 'minecraft:player' ? 1 : 0
                    if (!entity.typeId.includes('entity:')) isPlayer ? velocity ? entity.applyKnockback(k.x, k.z, k.hS, k.vS) : 0 : entity.applyImpulse(impulse)
                }
            }
        }
    } catch (e) { console.warn(); }
    try {
        if (s.id === 'launch:pad') {
            const entity = s.sourceEntity
            const isPlayer = entity.typeId === 'minecraft:player' ? 1 : 0
            const isItem = entity.typeId === 'minecraft:item' ? { y: 1, xz: 1 } : { y: 1.4, xz: 3 }
            switch (Number(s.message)) {
                case 0: isPlayer ? entity.applyKnockback(0, -1, -5, 1.4) : entity.applyImpulse({ x: 0, y: isItem.y, z: isItem.xz }); break
                case 1: isPlayer ? entity.applyKnockback(0, 1, -5, 1.4) : entity.applyImpulse({ x: 0, y: isItem.y, z: -isItem.xz }); break
                case 2: isPlayer ? entity.applyKnockback(-1, 0, -5, 1.4) : entity.applyImpulse({ x: isItem.xz, y: isItem.y, z: 0 }); break
                case 3: isPlayer ? entity.applyKnockback(1, 0, -5, 1.4) : entity.applyImpulse({ x: -isItem.xz, y: isItem.y, z: 0 }); break
            }
        }
    } catch (e) { }
})

world.afterEvents.entitySpawn.subscribe(({ entity }) => {
    try {
        if (entity.typeId === 'entity:quarry_function') {
            if (entity.getComponent('minecraft:mark_variant').value === 0) {
                quarryFunctions(entity, 1)
            } else if (entity.getComponent('minecraft:mark_variant').value === 1) {
                quarryFunctions(entity, 2)
            } else if (entity.getComponent('minecraft:mark_variant').value === 2) {
                quarryFunctions(entity, 3)
            }
        }
        else if (entity.typeId === 'minecraft:lightning_bolt') {
            let { x: cX, y: cY, z: cZ } = entity.location
            cX = Math.floor(cX); cY = Math.floor(cY); cZ = Math.floor(cZ)
            for (let x = cX - 8; x <= cX + 8; x++) {
                for (let y = cY - 8; y <= cY + 8; y++) {
                    for (let z = cZ - 8; z <= cZ + 8; z++) {
                        const distance = Math.sqrt((x - cX) ** 2 + (y - cY) ** 2 + (z - cZ) ** 2)
                        try {
                            const block = world.getDimension(entity.dimension.id).getBlock({ x, y, z })
                            if (distance <= 8) {
                                if (block.hasTag('cable')) {
                                    for (const tag of block.getTags()) {
                                        switch (tag) {
                                            case 'power_cable': block.dimension.runCommand(`setblock ${x} ${y} ${z} effect99:powered_cable`); break
                                            case 'power_cable_red': block.dimension.runCommand(`setblock ${x} ${y} ${z} effect99:powered_cable_red`); break
                                            case 'power_cable_blue': block.dimension.runCommand(`setblock ${x} ${y} ${z} effect99:powered_cable_blue`); break
                                            case 'power_cable_green': block.dimension.runCommand(`setblock ${x} ${y} ${z} effect99:powered_cable_green`); break
                                            case 'cable_controller': block.dimension.runCommand(`setblock ${x} ${y} ${z} effect99:powered_cable_controller`); break
                                            case 'reinforced_cable': block.dimension.runCommand(`setblock ${x} ${y} ${z} effect99:powered_reinforced_cable`); break
                                            case 'reinforced_cable_red': block.dimension.runCommand(`setblock ${x} ${y} ${z} effect99:powered_reinforced_cable_red`); break
                                            case 'reinforced_cable_blue': block.dimension.runCommand(`setblock ${x} ${y} ${z} effect99:powered_reinforced_cable_blue`); break
                                            case 'reinforced_cable_green': block.dimension.runCommand(`setblock ${x} ${y} ${z} effect99:powered_reinforced_cable_green`); break
                                        }
                                    }
                                }
                                else if (block.hasTag('battery')) {
                                    const permutation = block.permutation
                                    try { entity.runCommand(setPermutations(x, y, z, block.typeId, permutation.getAllStates(), { "energy:level": Math.floor(Math.random() * 2) ? 10 : 12 })) } catch (e) { }
                                }
                            }
                        } catch (e) { }
                    }
                }
            }
        }
    } catch (e) { }
})

world.afterEvents.pistonActivate.subscribe(p => {
    for (const { x, y, z } of p.piston.getAttachedBlocks()) {
        if (nonPushableBlocks.includes(p.dimension.getBlock({ x, y, z }).typeId)) {
            p.cancel = true
        }
        if (p.dimension.getBlock({ x, y, z }).typeId === 'effect99:quarry_drill') {
            const rotation = p.dimension.getBlock({ x, y, z }).permutation.getState('block:rotation')
            switch (rotation) {
                case 0: blockDestroy(x, y + 2, z, p.dimension.id, x, y, z, true, x, y + 1, z, rotation); break
                case 1: blockDestroy(x, y - 2, z, p.dimension.id, x, y, z, true, x, y - 1, z, rotation); break
                case 2: blockDestroy(x, y, z + 2, p.dimension.id, x, y, z, true, x, y, z + 1, rotation); break
                case 3: blockDestroy(x, y, z - 2, p.dimension.id, x, y, z, true, x, y, z - 1, rotation); break
                case 4: blockDestroy(x + 2, y, z, p.dimension.id, x, y, z, true, x + 1, y, z, rotation); break
                case 5: blockDestroy(x - 2, y, z, p.dimension.id, x, y, z, true, x - 1, y, z, rotation); break
            }
        }
    }
})

function blockDestroy(rx, ry, rz, dimension, x, y, z, piston = false, dx = rx, dy = ry, dz = rz, r = 0) {
    const block = world.getDimension(dimension).getBlock({ x, y, z }).permutation
    const typeId = world.getDimension(dimension).getBlock({ x: rx, y: ry, z: rz }).typeId
    if (!piston) {
        if (block.getState('block:energy')) {
            if (!nonBreakableBlocks.includes(typeId)) {
                world.getDimension(dimension).runCommand(setPermutations(x, y, z, block.type.id, block.getAllStates(), { "block:energy": block.getState('block:energy') - 1 }))
                world.getDimension(dimension).runCommand(`setblock ${rx} ${ry} ${rz} air 0 destroy`)
            }
        }
    } else {
        const ticks = (t = 2) => system.run(() => {
            try {
                const block = world.getDimension(dimension).getBlock({ x: dx, y: dy, z: dz }).permutation
                if (!t && r === world.getDimension(dimension).getBlock({ x: dx, y: dy, z: dz }).permutation.getState('block:rotation')) {
                    let permutations = { "block:cooldown": 2 }
                    if (block.getState('block:energy')) {
                        if (!nonBreakableBlocks.includes(typeId) && !typeId.includes('deepslate')) {
                            permutations["block:energy"] = block.getState('block:energy') - 1
                            world.getDimension(dimension).runCommand(`setblock ${rx} ${ry} ${rz} air 0 destroy`)
                        }
                    } else if (piston) {
                        if (!nonBreakableBlocks.includes(typeId) && !typeId.includes('deepslate')) {
                            world.getDimension(dimension).runCommand(`setblock ${rx} ${ry} ${rz} air 0 destroy`)
                        }
                    }
                    world.getDimension(dimension).runCommand(setPermutations(dx, dy, dz, block.type.id, block.getAllStates(), permutations))
                }
            } catch (e) { }
            if (t) ticks(t - 1)
        })
        ticks()
    }
}

const panelMenus = [
    'effect99:steam_generator',
    'effect99:steam_engine',
    'effect99:inefficient_steam_generator',
    'effect99:inefficient_steam_engine'
]

world.afterEvents.itemStartUseOn.subscribe(i => {
    const { x, y, z } = i.block
    const player = i.source
    const item = i.itemStack
    const dimension = player.dimension
    const block = i.block
    if (block.hasTag('security_door') && block.permutation.getState('door:top_bottom') === 2 && !block.permutation.getState('block:model') && !dimension.getBlock({ x: x, y: y - 1, z: z }).permutation.getState('block:model')) {
        let entity = 0
        for (const password of dimension.getEntitiesAtBlockLocation({ x, y, z })) {
            if (password.typeId === 'entity:security_door') {
                entity = 1
                settPassword(password)
            }
        }
        if (!entity) {
            settPassword(player.dimension.spawnEntity('entity:security_door', { x: x + 0.5, y: y + 0.5, z: z + 0.5 }))
        }
        function settPassword(passW) {
            !passW.getTags()[0] ? passW.addTag(JSON.stringify({ passW: '', openOwner: 0, autoClose: 0, destroy: 0, owner: '' })) : 0
            const sett = JSON.parse(passW.getTags()[0])
            const menu = new ModalFormData().title('Security Door')
                .textField(`${sett.passW ? `Owner: ${sett.owner}` : 'No password, please enter a new password'}`, '§7Enter the password')
                .toggle('Skip owner', sett.openOwner ? true : false)
                .toggle('Close automatically', sett.autoClose ? true : false)
                .toggle('Destroy door')

            menu.show(player).then(r => {
                if (!r.canceled) {
                    if (sett.passW) {
                        const owner = sett.owner
                        const pass = sett.passW
                        if (!sett.openOwner || sett.passW === `${r.formValues[0]}`) {
                            if (sett.passW === `${r.formValues[0]}`) {
                                passW.removeTag(`${passW.getTags()[0]}`)
                                passW.addTag(JSON.stringify({ passW: `${r.formValues[0]}`, openOwner: r.formValues[1] ? 1 : 0, autoClose: r.formValues[2] ? 1 : 0, destroy: r.formValues[3] ? 1 : 0, owner: `${owner}` }))
                                if (r.formValues[3]) dimension.fillBlocks({ x, y, z }, { x, y, z }, BlockTypes.get('minecraft:air'))
                                else {
                                    player.dimension.spawnEntity('entity:door_signal<openDoor>', { x: x, y: y - 1, z: z })
                                    if (r.formValues[2]) passW.triggerEvent('autoClose')
                                }
                            } else {
                                player.onScreenDisplay.setActionBar(`§4Incorrect password\n§f${r.formValues[0]}`)
                            }
                        }
                        else if (owner === player.name) {
                            passW.removeTag(`${passW.getTags()[0]}`)
                            passW.addTag(JSON.stringify({ passW: `${pass}`, openOwner: r.formValues[1] ? 1 : 0, autoClose: r.formValues[2] ? 1 : 0, destroy: r.formValues[3] ? 1 : 0, owner: `${owner}` }))
                            if (r.formValues[3]) dimension.fillBlocks({ x, y, z }, { x, y, z }, BlockTypes.get('minecraft:air'))
                            else {
                                player.dimension.spawnEntity('entity:door_signal<openDoor>', { x: x, y: y - 1, z: z })
                                if (r.formValues[2]) passW.triggerEvent('autoClose')
                            }
                        } else {
                            player.onScreenDisplay.setActionBar(`§cYou are not the owner of the door`)
                        }
                    } else {
                        passW.removeTag(`${passW.getTags()[0]}`)
                        passW.addTag(JSON.stringify({ passW: `${r.formValues[0]}`, openOwner: r.formValues[1] ? 1 : 0, autoClose: r.formValues[2] ? 1 : 0, destroy: r.formValues[3] ? 1 : 0, owner: `${player.name}` }))
                        if (r.formValues[3]) dimension.fillBlocks({ x, y, z }, { x, y, z }, BlockTypes.get('minecraft:air'))
                        else {
                            player.dimension.spawnEntity('entity:door_signal<openDoor>', { x: x, y: y - 1, z: z })
                            if (r.formValues[2]) passW.triggerEvent('autoClose')
                        }
                    }
                }
            })
        }
    }
    else if (block.hasTag('conveyor_belt_filter')) {
        let entity = 0
        for (const filter of dimension.getEntitiesAtBlockLocation({ x, y, z })) {
            if (filter.typeId === 'entity:conveyor_belt_filter') {
                entity = 1
                itemFilter(filter, player)
            }
        }
        if (!entity) {
            itemFilter(player.dimension.spawnEntity('entity:conveyor_belt_filter', { x: x + 0.5, y: y + 0.5, z: z + 0.5 }), player)
        }
        function itemFilter(f, player) {
            const menu = new ModalFormData().title('Conveyor belt filter')
                .textField(`§8Filtered items: ${f.getTags().length} ${item ? `\n§6Item in your hand: ${item.typeId}` : ''} \n§7Item Filter: Item ID`, 'E.g: minecraft:coal, minecraft:redstone', `${f.getTags()[0] ? f.getTags() : item ? item.typeId : ''}`)
                .toggle('Floating text', f.nameTag ? (f.getTags()[0] ? true : false) : false).toggle('Inverted filter', block.permutation.getState('filter:inverted') ? true : false)
            menu.show(player).then(r => {
                if (!r.canceled) {
                    try {
                        for (const item of f.getTags()) {
                            f.removeTag(item)
                        }
                        for (const item of r.formValues[0].split(',')) {
                            if (!item.includes(':')) continue
                            f.addTag(item)
                        }
                    } catch (e) { }
                    let t = f.getTags()[0]
                    r.formValues[1] && t ? f.nameTag = `${t.slice(t.indexOf(":") + 1)[0].toUpperCase() + t.slice(t.indexOf(":") + 2).replace(/_/g, " ")}` : f.nameTag = ''
                    const permutation = block.permutation
                    if (r.formValues[2]) {
                        f.runCommand(setPermutations(x, y, z, block.typeId, permutation.getAllStates(), { "filter:inverted": 1 }))
                    } else {
                        f.runCommand(setPermutations(x, y, z, block.typeId, permutation.getAllStates(), { "filter:inverted": t ? 0 : 1 }))
                    }
                }
            })
        }
    }
    else if (block.hasTag('storage_core_panel')) {
        let coords;
        switch (block.permutation.getState('block:rotation')) {
            case 0: coords = { x, y: y + 1, z: z - 2 }; break
            case 1: coords = { x, y: y + 1, z: z + 2 }; break
            case 2: coords = { x: x - 2, y: y + 1, z }; break
            case 3: coords = { x: x + 2, y: y + 1, z }; break
        }
        const { x: sx, y: sy, z: sz } = coords
        let num = 0; let math = 0; let core = 0; let error = 1
        let entity
        for (entity of dimension.getEntitiesAtBlockLocation({ x: sx, y: sy, z: sz })) {
            if (entity.typeId === 'entity:storage_core') {
                try {
                    num = entity.getTags()[0] ? entity.getTags().map(item => JSON.parse(item).amount).reduce((add, item) => add + item, 0) : 0
                    math = num - 999
                    math = math > 0 && math <= 100000 ? Math.ceil(math / 1000) : '§8Less than one percent'
                } catch (e) { }
                core = entity
                error = 0
            }
        }
        const panel = new ActionFormData().title('Storage Core')
            .body(`Stored items: ${num}§7/100.000\nPct Approx of items stored: ${math}‰\n\n§rMachine energy level: ${Math.ceil(block.permutation.getState('energy:level') / 0.08)}‰\n\n${error ? '§4Error: internal memory of storage core not detected, please dismantle the machine and rebuild it.\n\n' : ''}`)
                .button(`Storage Core Settings`, 'textures/ui/settings_glyph_color_2x').button(`Manage Storage`, 'textures/ui/storageIconColor')

        panel.show(player).then(r => {
            if (r.selection === 0) {
                const settingsPanel = new ModalFormData().title('Storage Core Settings')
                    .dropdown(`Storage Core mode \n§8This will only affect this side of the machine`, ['Storage mode', 'Emptying mode'], block.permutation.getState('storage:mode') ? true : false)
                    .slider('Maximum Number of Item Output', 1, 64, 1, core ? core.getComponent('minecraft:mark_variant').value : 64)
                    .toggle('Dismantle Machine? §7Warning, by destroying the machine you will lose all stored items.')
                settingsPanel.show(player).then(r => {
                    if (!r.canceled) {
                        const permutation = block.permutation
                        dimension.runCommand(setPermutations(x, y, z, block.typeId, permutation.getAllStates(), { "storage:mode": r.formValues[0] }))
                        if (core) core.getComponent('minecraft:mark_variant').value = r.formValues[1]
                        if (r.formValues[2]) {
                            dimension.runCommand(`fill ${sx + 2} ${sy - 2} ${sz + 2} ${sx - 2} ${sy} ${sz - 2} air 0`)
                            dimension.spawnItem(new ItemStack('effect99:storage_core_placer', 1), { x: sx, y: sy - 1, z: sz })
                        }
                    }
                })
            } else if (r.selection === 1 && entity) {
                allItemList(player, entity)
            }
        })
    }
    else if (block.hasTag('machine_panel') && panelMenus.includes(block.typeId)) {
        const panel = new ActionFormData().title(`${block.typeId.includes('engine') ? 'Steam Engine' : 'Steam Generator'}`).body('Dismantle Machine?').button('Yes').button('No')
        panel.show(player).then(r => {
            if (r.selection === 0) {
                const air = BlockTypes.get('minecraft:air')
                if (block.typeId.includes('steam_engine') && !block.typeId.includes('inefficient')) { dimension.spawnItem(new ItemStack('effect99:steam_engine_placer'), { x, y, z }) }
                else if (block.typeId.includes('steam_generator') && !block.typeId.includes('inefficient')) { dimension.spawnItem(new ItemStack('effect99:steam_generator_placer'), { x, y, z }) }
                else if (block.typeId.includes('steam_engine') && block.typeId.includes('inefficient')) { dimension.spawnItem(new ItemStack('effect99:inefficient_steam_engine_placer'), { x, y, z }) }
                else if (block.typeId.includes('steam_generator') && block.typeId.includes('inefficient')) { dimension.spawnItem(new ItemStack('effect99:inefficient_steam_generator_placer'), { x, y, z }) }

                if (block.hasTag('north_block')) {
                    !block.typeId.includes('inefficient') ? dimension.fillBlocks({ x: x - 1, y: y - 1, z: z - 2 }, { x: x + 1, y: y + 1, z }, air) : dimension.fillBlocks({ x: x - 1, y: y - 1, z: z - 2 }, { x: x + 1, y, z }, air)
                }
                else if (block.hasTag('south_block')) {
                    !block.typeId.includes('inefficient') ? dimension.fillBlocks({ x: x + 1, y: y + 1, z: z + 2 }, { x: x - 1, y: y - 1, z }, air) : dimension.fillBlocks({ x: x + 1, y, z: z + 2 }, { x: x - 1, y: y - 1, z }, air)
                }
                else if (block.hasTag('west_block')) {
                    !block.typeId.includes('inefficient') ? dimension.fillBlocks({ x: x - 2, y: y - 1, z: z - 1 }, { x, y: y + 1, z: z + 1 }, air) : dimension.fillBlocks({ x: x - 2, y: y - 1, z: z - 1 }, { x, y, z: z + 1 }, air)
                }
                else if (block.hasTag('east_block')) {
                    !block.typeId.includes('inefficient') ? dimension.fillBlocks({ x: x + 2, y: y + 1, z: z + 1 }, { x, y: y - 1, z: z - 1 }, air) : dimension.fillBlocks({ x: x + 2, y, z: z + 1 }, { x, y: y - 1, z: z - 1 }, air)
                }
            }
        })
    }
})

function allItemList(player, entity) {
    entity.triggerEvent('StopMachine')
    const managePanel = new ActionFormData().title('Storage of the Core').button(`§lSearch`, 'textures/ui/magnifyingGlass')
    const itemSaved = entity.getTags().map(item => JSON.parse(item)).reverse()

    itemSaved.forEach(({ typeId, amount }, i) => {
        managePanel.button(`${i % 2 === 0 ? '§e' : '§9'}${typeId[typeId.indexOf(":") + 1].toUpperCase() + typeId.slice(typeId.indexOf(":") + 2).replace(/_/g, " ")} §j${amount}`,
            `textures/ui/${iconBox[Math.floor(Math.random() * iconBox.length)]}`)
    })
    managePanel.body(`§8Slots: ${itemSaved.length}`)
    managePanel.show(player).then(r => {
        if (r.selection === 0) {
            searchItem(player, entity)

        } else if (r.selection) {
            giveItems(itemSaved, player, entity, r)
        }
        else if (r.canceled) {
            entity.triggerEvent('default')
        }
    })
}

function giveItems(itemSaved, player, entity, r, search) {
    const item = itemSaved[r.selection - 1]
    const { typeId, amount, dur, ench } = item
    const itemStack = new ItemStack(typeId)
    const durability = itemStack.getComponent("durability") ? itemStack.getComponent("minecraft:durability").maxDurability : undefined
    const itemSelected = new ModalFormData().title(`Give Item`)
        .slider(`${typeId[typeId.indexOf(":") + 1].toUpperCase() + typeId.slice(typeId.indexOf(":") + 2).replace(/_/g, " ")}
        \n§‰§pAmount: ${amount} ${durability ? `\nDurability ${dur}/${durability}` : typeId === 'effect99:item_hoover' ? `\nBattery ${dur}‰` : ''}
        ${ench[0] ? '\n' + ench.reduce((add, { enchant, level }) => add + (`\n${enchant[enchant.indexOf(":") + 1].toUpperCase() + enchant.slice(enchant.indexOf(":") + 2).replace(/_/g, " ")} ${level}`), '').trim() : ''}
        \n§7Count`, 1, amount, 1, Math.min(itemStack.maxAmount, amount))

    itemSelected.show(player).then(is => {
        if (!is.canceled) {
            const { x, y, z } = player.location
            let numCount = is.formValues[0]
            let max = itemStack.maxAmount
            let index = r.selection - 1
            let result = []
            while (numCount > 0) { result.push(Math.min(max, numCount)); numCount -= max }
            result.forEach(stacks => {
                let itemSaved = []
                search ? itemSaved = entity.getTags().map(item => JSON.parse(item)).filter(item => item.typeId.includes(search)).reverse()
                    : itemSaved = entity.getTags().map(item => JSON.parse(item)).reverse()

                const item = itemSaved[index]
                const { typeId, amount, dur, ench } = item
                const itemStack = new ItemStack(typeId, stacks)
                const lore = typeId === 'effect99:item_hoover' ? dur : 0
                dur && !lore ? itemStack.getComponent("durability").damage = Math.min(dur, durability) : 0
                const enchantments = itemStack.getComponent("enchantable")
                for (const { enchant, level } of ench) {
                    enchantments.addEnchantment({type: enchant, level: level})
                }
                if (lore) itemStack.setLore([`§r§lBattery ${dur}‰`, '§r§l§cOff'])
                if (amount === 1 || amount - stacks === 0) {
                    entity.removeTag(JSON.stringify(item))
                } else {
                    entity.removeTag(JSON.stringify(item))
                    entity.addTag(JSON.stringify({ typeId, amount: amount - stacks, dur, ench }))
                    search ? itemSaved = entity.getTags().map(item => JSON.parse(item)).filter(item => item.typeId.includes(search)).reverse() : itemSaved = entity.getTags().map(item => JSON.parse(item)).reverse()
                    itemSaved.forEach((object, i) => {
                        if (matchedObject([{ typeId, amount: amount - stacks, dur, ench }], [object])) {
                            index = i
                        }
                    })
                }
                player.dimension.spawnItem(itemStack, { x, y: y + 0.5, z })
            })
            entity.triggerEvent('default')
        } 
        else {
            entity.triggerEvent('default')
        }
    })
}

function searchItem(player, entity) {
    const itemSearch = new ModalFormData().title(`Search Item`).textField(' ', 'Search: e.g stone')
    itemSearch.show(player).then(r => {
        if (!r.canceled) {
            const managePanel = new ActionFormData().title('Storage of the Core').button(`§lSearch`, 'textures/ui/magnifyingGlass')
            const itemSaved = entity.getTags().map(item => JSON.parse(item)).filter(item => item.typeId.includes(r.formValues[0].toLowerCase())).reverse()
            managePanel.body(`§7Results: ${itemSaved.length}`)
            itemSaved.forEach(({ typeId, amount }, i) => {
                managePanel.button(`${i % 2 === 0 ? '§e' : '§9'}${typeId[typeId.indexOf(":") + 1].toUpperCase() + typeId.slice(typeId.indexOf(":") + 2).replace(/_/g, " ")} §j${amount}`,
                    `textures/ui/${iconBox[Math.floor(Math.random() * iconBox.length)]}`)
            })
            managePanel.button(`All Items`)
            managePanel.show(player).then(s => {
                if (s.selection === 0) {
                    searchItem(player, entity)
                }
                else if (s.selection && s.selection != itemSaved.length + 1) {
                    giveItems(itemSaved, player, entity, s, r.formValues[0].toLowerCase())
                }
                else if (s.selection === itemSaved.length + 1) {
                    allItemList(player, entity)
                }
                else if (s.canceled) {
                    entity.triggerEvent('default')
                }
            })
        } else {
            entity.triggerEvent('default')
        }
    })
}

function matchedObject(a, b) {
    if (a.length !== b.length) {
        return false;
    }
    return a.every(item => b.some(
        (otherItem) => JSON.stringify(item) === JSON.stringify(otherItem)
    ))
}

function quarryFunctions(e, r) {
    let warning = false
    let { x, y, z } = e.location
    x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
    for (let i = -r; i <= r; i++) {
        for (let k = -r; k <= r; k++) {
            try {
                const block = e.dimension.getBlock({ x: x + i, y: y, z: z + k })
                if (qNonBreakableBlocks.includes(block.typeId)) { // warning
                    for (let l = 0; l < 384; l++) {
                        const b = e.dimension.getBlock({ x: x, y: y + l, z: z })
                        if (b.hasTag('quarry_center') && !b.hasTag('power_on_0')) {
                            warning = true
                            const permutation = e.dimension.getBlock({ x: x, y: y + l, z: z }).permutation
                            e.dimension.runCommand(setPermutations(x, y + l, z, permutation.type.id, permutation.getAllStates(), { "cooldown:time": 3 }))
                            e.triggerEvent('despawn')
                            break
                        }
                        else if (b.hasTag('warning')) {
                            warning = true
                            break
                        }
                    }
                    break
                }
            } catch (e) { }
        }
    }
    try {
        const block = e.dimension.getBlock({ x: x, y: y + 1, z: z })
        if (!block.hasTag('warning') && block.hasTag('quarry_structure_drill')) { //fill
            for (let l = 0; l < 384; l++) {
                const b = e.dimension.getBlock({ x: x, y: y + l, z: z })
                if (y + l === 383) break
                else if (b.hasTag('quarry_center') && !block.hasTag('warning') && !b.hasTag('power_on_0') && !b.hasTag('power_on_1') && !b.hasTag('quarry_center_off') && !warning) {
                    e.triggerEvent('fill_cooldown')
                    const permutation = e.dimension.getBlock({ x: x, y: y + l, z: z }).permutation
                    e.dimension.runCommand(setPermutations(x, y + l, z, permutation.type.id, permutation.getAllStates(), { "power:on": permutation.getState('power:on') - 1 }))
                    break
                } else if (b.hasTag('quarry_center')) {
                    e.triggerEvent('despawn')
                    break
                }
            }
        } else if (block.hasTag('warning')) { //valid
            try {
                for (let i = -r; i <= r; i++) {
                    for (let k = -r; k <= r; k++) {
                        const block = e.dimension.getBlock({ x: x + i, y: y, z: z + k })
                        if (qNonBreakableBlocks.includes(block.typeId)) warning = true
                    }
                }
                if (!warning) {
                    for (let l = 0; l < 384; l++) {
                        const b = e.dimension.getBlock({ x: x, y: y + l, z: z })
                        if (y + l === 383) break
                        else if (b.hasTag('quarry_center') && block.hasTag('warning') && !b.hasTag('power_on_0')) {
                            e.triggerEvent('fill_cooldown')
                            const permutation = e.dimension.getBlock({ x: x, y: y + l, z: z }).permutation
                            e.dimension.runCommand(setPermutations(x, y + l, z, permutation.type.id, permutation.getAllStates(), { "cooldown:time": 1 }))
                            e.triggerEvent('despawn')
                            break
                        }
                    }
                }
            } catch (e) { }
        }
    } catch (e) { }
}

const iconBox = ['box_yellow', 'box_green', 'box_blue', 'box_pink', 'box_orange', 'box_gray']

const nonPushableBlocks = [
    'effect99:mechanical_arm',
    'effect99:fast_mechanical_arm',
    'effect99:regular_mechanical_arm',
    'effect99:antenna_base',
    'effect99:antenna',
    'effect99:antenna_signal',
    'effect99:steam_engine',
    'effect99:steam_generator',
    'effect99:quarry_structure',
    'effect99:medium_quarry_structure',
    'effect99:largue_quarry_structure',
    'effect99:small_quarry',
    'effect99:medium_quarry',
    'effect99:largue_quarry',
    'effect99:storage_core',
    'effect99:isg_blocks'
]

const qNonBreakableBlocks = [
    'minecraft:bedrock',
    'minecraft:end_portal_frame',
    'minecraft:lava',
    'minecraft:water',
    'minecraft:flowing_lava',
    'minecraft:flowing_water',
    'effect99:steam_engine',
    'effect99:steam_generator',
    'effect99:quarry_structure',
    'effect99:medium_quarry_structure',
    'effect99:largue_quarry_structure',
    'minecraft:mob_spawner',
    'effect99:small_quarry',
    'effect99:medium_quarry',
    'effect99:largue_quarry',
    'effect99:storage_core',
    'effect99:isg_blocks'
]

const nonBreakableBlocks = [
    'minecraft:air',
    'minecraft:lava',
    'minecraft:water',
    'minecraft:bedrock',
    'minecraft:obsidian',
    'effect99:isg_blocks',
    'effect99:quarry_drill',
    'effect99:steam_engine',
    'minecraft:moving_block',
    'minecraft:flowing_lava',
    'minecraft:flowing_water',
    'effect99:chunks_viewer',
    'effect99:small_quarry',
    'effect99:medium_quarry',
    'effect99:largue_quarry',
    'effect99:storage_core',
    'effect99:steam_generator',
    'minecraft:ancient_debris',
    'minecraft:crying_obsidian',
    'effect99:quarry_structure',
    'minecraft:end_portal_frame',
    'minecraft:reinforced_deepslate',
    'minecraft:piston_arm_collision',
    'effect99:medium_quarry_structure',
    'effect99:largue_quarry_structure',
    'minecraft:sticky_piston_arm_collision'
]

world.afterEvents.worldInitialize.subscribe(() => {
    world.getDimension('overworld').runCommand('gamerule sendcommandfeedback false')
    world.getDimension('overworld').runCommand('gamerule commandblockoutput false')
})