import { world, ItemStack, system, Vector } from "@minecraft/server"

const dimensions = ['overworld', 'nether', 'the_end']
let admin = false
let itemMovementTicks = 0
let itemsInMovement = []
let worldEntities = []

function setPermutations(x, y, z, blockName, properties, values) {
    const props = { ...properties };
    for (const [key, val] of Object.entries(values)) {
        if (props.hasOwnProperty(key)) {
            props[key] = val;
        }
    }
    const propStrings = Object.entries(props).map(([key, val]) => `"${key}": ${val}`);
    return `setblock ${x} ${y} ${z} ${blockName} [ ${propStrings.join(", ")} ]`;
}


const worldTicks = () => system.run(() => {
    const t = system.currentTick
    if (t % 4 === 0) {
        for (const player of world.getAllPlayers()) {
            const item = player.getComponent("inventory").container.getItem(player.selectedSlot)
            const { x, y, z } = player.location
            try {
                const block = player.dimension.getBlock({ x: Math.floor(x), y: Math.floor(y + 1.5), z: Math.floor(z) }).typeId
                if (block === 'effect99:conveyor_tube_up' || block === 'effect99:conveyor_tube_down' || block === 'effect99:conveyor_tube_station' || block === 'effect99:conveyor_tube_station_up' || block === 'effect99:conveyor_tube_up_off' || block === 'effect99:conveyor_tube_down_off') {
                    player.runCommand('playanimation @s animation.conveyor.tube')
                } else if (block === 'effect99:conveyor_tube_north' || block === 'effect99:conveyor_tube_south' || block === 'effect99:conveyor_tube_east' || block === 'effect99:conveyor_tube_west' || block === 'effect99:conveyor_tube_north_off' || block === 'effect99:conveyor_tube_south_off' || block === 'effect99:conveyor_tube_east_off' || block === 'effect99:conveyor_tube_west_off') {
                    player.runCommand('playanimation @s animation.conveyor.tube_sides')
                }
                if (player.isSneaking && player.dimension.getBlock({ x: Math.floor(x), y: Math.floor(y - 1), z: Math.floor(z) }).hasTag('not_full_storage') && player.level >= 1) {
                    const permutation = player.dimension.getBlock({ x: Math.floor(x), y: Math.floor(y - 1), z: Math.floor(z) }).permutation
                    player.runCommand(setPermutations(Math.floor(x), Math.floor(y - 1), Math.floor(z), permutation.type.id, permutation.getAllStates(), { "xp:level": permutation.getState('xp:level') + 1 }))
                    player.runCommand('xp -1l @s')
                }
            } catch (e) { }
            try {
                if (item.typeId === 'effect99:item_hoover') {
                    if (item.getLore()[0]) {
                        if (Number(item.getLore()[0].match(/[\d.]+/g)[0])) {
                            player.onScreenDisplay.setActionBar(item.getLore().reduce((x, y) => (x + '\n     ' + y)))
                        }
                    } else {
                        item.setLore(['§r§lBattery 0‰', '§r§l§cOff'])
                        player.getComponent('inventory').container.setItem(player.selectedSlot, item)
                    }
                    const energy = Number(item.getLore()[0].match(/[\d.]+/g)[0])
                    if (item.getLore()[1] === ('§r§l§aOn')) {
                        let inUse = false
                        let maxPower = false
                        const { x: vx, y: vy, z: vz } = player.getBlockFromViewDirection({ maxDistance: 12, includeLiquidBlocks: true }).block

                        for (const entity of player.dimension.getEntities({ location: { x: vx, y: vy + 1, z: vz }, maxDistance: 8 })) {
                            if (entity.getComponent('minecraft:item')) {
                                inUse = true
                                const { x: ex, y: ey, z: ez } = entity.location
                                const velocity = new Vector(x - ex, y + 1 - ey, z - ez).normalized()
                                if (entity.dimension.getBlock({ x: ex, y: ey - 1, z: ez }).typeId === 'minecraft:air') {
                                    try { entity.applyImpulse({ x: velocity.x * 0.5, y: velocity.y * 0.5, z: velocity.z * 0.5 }) } catch (e) { }
                                } else {
                                    try { entity.applyImpulse({ x: velocity.x * 0.38, y: velocity.y * 0.38, z: velocity.z * 0.38 }) } catch (e) { }
                                }
                            }
                        }

                        for (let i = -2; i <= 2; i++) {
                            for (let j = -1; j <= 1; j++) {
                                for (let k = -2; k <= 2; k++) {
                                    const block = player.dimension.getBlock({ x: vx + i, y: vy + j, z: vz + k });
                                    try {
                                        const permutation = block.permutation
                                        if (permutation.getState('growth') === 7) {
                                            block.dimension.runCommand(`setblock ${vx + i} ${vy + j} ${vz + k} ${block.typeId} 0 destroy`)
                                            inUse = true
                                            break
                                        }
                                    } catch (e) { }
                                    if (block.typeId === 'minecraft:sand' || block.typeId === 'minecraft:gravel') {
                                        block.dimension.runCommand(`setblock ${vx + i} ${vy + j} ${vz + k} air 0 destroy`)
                                        inUse = true
                                        maxPower = true
                                        break
                                    }
                                }
                            }
                        }
                        if (inUse) {
                            if (energy) {
                                item.setLore([`§r§lBattery ${(energy >= 1 ? maxPower ? energy - 1 : energy - 0.1 : energy - 0.1).toFixed(1)}‰`, '§r§l§aOn'])
                                player.getComponent('inventory').container.setItem(player.selectedSlot, item)
                            } else {
                                item.setLore([`§r§lBattery 0‰`, '§r§l§cOff'])
                                player.getComponent('inventory').container.setItem(player.selectedSlot, item)
                            }
                        }
                    } else if (energy === 0) {
                        let battery = 0
                        for (let i = 0; i < 36; i++) {
                            const inventory = player.getComponent('inventory').container
                            const { typeId, amount } = inventory.getItem(i) ? inventory.getItem(i) : { typeId: '', amount: 0 }
                            switch (typeId) {
                                case 'effect99:small_battery_i': recharge(10); break
                                case 'effect99:small_battery_ii': recharge(20); break
                                case 'effect99:small_battery_iii': recharge(30); break
                                case 'effect99:small_battery_iv': recharge(40); break
                                case 'effect99:small_battery_v': recharge(50); break
                                case 'effect99:small_battery_vi': recharge(60); break
                                case 'effect99:small_battery_vii': recharge(70); break
                                case 'effect99:small_battery_viii': recharge(80); break
                                case 'effect99:small_battery_ix': recharge(90); break
                                case 'effect99:small_battery_x': recharge(100); break
                            }
                            function recharge(charge) {
                                if (amount > 1) {
                                    inventory.setItem(i, new ItemStack(typeId, amount - 1))
                                }
                                let full = 1
                                for (let j = 0; j < 36; j++) {
                                    const { typeId, amount: count } = inventory.getItem(j) ? inventory.getItem(j) : { typeId: '', amount: 0 }
                                    if (amount > 1 && !typeId) {
                                        inventory.setItem(j, new ItemStack('effect99:battery', 1))
                                        full = 0; break
                                    }
                                    else if (amount > 1 && typeId === 'effect99:battery' && count != 64) {
                                        inventory.setItem(j, new ItemStack('effect99:battery', count + 1))
                                        full = 0; break
                                    }
                                    else if (amount <= 1) {
                                        inventory.setItem(i, new ItemStack('effect99:battery'))
                                        full = 0; break
                                    }
                                }
                                if (full) player.dimension.spawnItem(new ItemStack('effect99:battery', 1), { x: x, y: y + 0.5, z: z })
                                item.setLore([`§r§lBattery ${charge}‰`, '§r§l§cOff'])
                                player.getComponent('inventory').container.setItem(player.selectedSlot, item)
                                battery = 1
                            }
                        }
                        if (!battery) player.onScreenDisplay.setActionBar('§l§‰Place a small\nbattery in inventory')
                    }
                }
            } catch (e) { }
            try {
                const itemId = item.typeId
                if (item && !admin && itemId.includes('effect99')) {
                    if (itemId === 'effect99:powered_cable') give(player, 'effect99:power_cable')
                    else if (itemId === 'effect99:powered_cable_red') give(player, 'effect99:power_cable_red')
                    else if (itemId === 'effect99:powered_cable_blue') give(player, 'effect99:power_cable_blue')
                    else if (itemId === 'effect99:powered_cable_green') give(player, 'effect99:power_cable_green')
                    else if (itemId === 'effect99:powered_reinforced_cable') give(player, 'effect99:reinforced_cable')
                    else if (itemId === 'effect99:powered_reinforced_cable_red') give(player, 'effect99:reinforced_cable_red')
                    else if (itemId === 'effect99:powered_reinforced_cable_blue') give(player, 'effect99:reinforced_cable_blue')
                    else if (itemId === 'effect99:powered_reinforced_cable_green') give(player, 'effect99:reinforced_cable_green')
                    else if (itemId === 'effect99:powered_cable_controller') give(player, 'effect99:power_cable_controller')
                    else if (itemId === 'effect99:quarry_structure' || itemId === 'effect99:small_quarry') give(player, 'effect99:small_quarry_in_box')
                    else if (itemId === 'effect99:medium_quarry_structure' || itemId === 'effect99:medium_quarry') give(player, 'effect99:medium_quarry_in_box')
                    else if (itemId === 'effect99:largue_quarry_structure' || itemId === 'effect99:largue_quarry') give(player, 'effect99:largue_quarry_in_box')
                    else if (itemId === 'effect99:steam_engine') give(player, 'effect99:steam_engine_placer')
                    else if (itemId === 'effect99:steam_generator') give(player, 'effect99:steam_generator_placer')
                    else if (itemId === 'effect99:inefficient_steam_engine') give(player, 'effect99:inefficient_steam_engine_placer')
                    else if (itemId === 'effect99:inefficient_steam_generator' || itemId === 'effect99:isg_blocks') give(player, 'effect99:inefficient_steam_generator_placer')
                    else if (itemId === 'effect99:storage_core') give(player, 'effect99:storage_core_placer')
                    else if (itemId === 'effect99:conveyor_belt_north' || itemId === 'effect99:conveyor_belt_south' || itemId === 'effect99:conveyor_belt_west' || itemId === 'effect99:conveyor_belt_east' || itemId === 'effect99:conveyor_belt_north_off' || itemId === 'effect99:conveyor_belt_south_off' || itemId === 'effect99:conveyor_belt_west_off' || itemId === 'effect99:conveyor_belt_east_off') give(player, 'effect99:conveyor_belt')
                    else if (itemId === 'effect99:conveyor_tube_north' || itemId === 'effect99:conveyor_tube_south' || itemId === 'effect99:conveyor_tube_west' || itemId === 'effect99:conveyor_tube_east' || itemId === 'effect99:conveyor_tube_up' || itemId === 'effect99:conveyor_tube_down' || itemId === 'effect99:conveyor_tube_north_off' || itemId === 'effect99:conveyor_tube_south_off' || itemId === 'effect99:conveyor_tube_west_off' || itemId === 'effect99:conveyor_tube_east_off' || itemId === 'effect99:conveyor_tube_up_off' || itemId === 'effect99:conveyor_tube_down_off') give(player, 'effect99:conveyor_tube')
                    else if (itemId === "effect99:antenna" || itemId === "effect99:antenna_signal" || itemId === "effect99:conveyor_tube_station_up" || itemId === "effect99:power_generating_furnace_up") give(player, 'minecraft:air', true)
                }
            } catch (e) { }
        }
        for (const worlds of dimensions) {
            try {
                for (const entity of world.getDimension(worlds).getEntities({ excludeTypes: ['minecraft:item'] })) {
                    const dimension = entity.dimension
                    let { x, y, z } = entity.location
                    x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                    if (entity.typeId === 'entity:mechanical_arm' || entity.typeId === 'entity:regular_mechanical_arm' || entity.typeId === 'entity:fast_mechanical_arm') {
                        const block = dimension.getBlock({ x, y, z }).permutation
                        switch (entity.getComponent('minecraft:variant').value) {
                            case 0:
                                if (dimension.getBlock({ x: x, y: y, z: z - 1 }).hasTag('need_fuel_1') && !block.getState('cooldown:arm')) fuelProvider(block, x, y, z, entity, x, y, z + 1, dimension, x, y, z - 1)
                                else if (dimension.getBlock({ x: x, y: y, z: z - 2 }).hasTag('need_fuel_2') && !block.getState('cooldown:arm')) fuelProvider(block, x, y, z, entity, x, y, z + 1, dimension, x, y, z - 2)
                                else if (!block.getState('cooldown:arm') && !dimension.getBlock({ x: x, y: y, z: z - 1 }).hasTag('adaptable_arm_1') && !dimension.getBlock({ x: x, y: y, z: z - 2 }).hasTag('adaptable_arm_2')) itemMovement(0, dimension, x, y, z + 1, block, x, y, z, entity)
                                break
                            case 1:
                                if (dimension.getBlock({ x: x, y: y, z: z + 1 }).hasTag('need_fuel_1') && !block.getState('cooldown:arm')) fuelProvider(block, x, y, z, entity, x, y, z - 1, dimension, x, y, z + 1)
                                else if (dimension.getBlock({ x: x, y: y, z: z + 2 }).hasTag('need_fuel_2') && !block.getState('cooldown:arm')) fuelProvider(block, x, y, z, entity, x, y, z - 1, dimension, x, y, z + 2)
                                else if (!block.getState('cooldown:arm') && !dimension.getBlock({ x: x, y: y, z: z + 1 }).hasTag('adaptable_arm_1') && !dimension.getBlock({ x: x, y: y, z: z + 2 }).hasTag('adaptable_arm_2')) itemMovement(1, dimension, x, y, z - 1, block, x, y, z, entity)
                                break
                            case 2:
                                if (dimension.getBlock({ x: x - 1, y: y, z: z }).hasTag('need_fuel_1') && !block.getState('cooldown:arm')) fuelProvider(block, x, y, z, entity, x + 1, y, z, dimension, x - 1, y, z)
                                else if (dimension.getBlock({ x: x - 2, y: y, z: z }).hasTag('need_fuel_2') && !block.getState('cooldown:arm')) fuelProvider(block, x, y, z, entity, x + 1, y, z, dimension, x - 2, y, z)
                                else if (!block.getState('cooldown:arm') && !dimension.getBlock({ x: x - 1, y: y, z: z }).hasTag('adaptable_arm_1') && !dimension.getBlock({ x: x - 2, y: y, z: z }).hasTag('adaptable_arm_2')) itemMovement(2, dimension, x + 1, y, z, block, x, y, z, entity)
                                break
                            case 3:
                                if (dimension.getBlock({ x: x + 1, y: y, z: z }).hasTag('need_fuel_1') && !block.getState('cooldown:arm')) fuelProvider(block, x, y, z, entity, x - 1, y, z, dimension, x + 1, y, z)
                                else if (dimension.getBlock({ x: x + 2, y: y, z: z }).hasTag('need_fuel_2') && !block.getState('cooldown:arm')) fuelProvider(block, x, y, z, entity, x - 1, y, z, dimension, x + 2, y, z)
                                else if (!block.getState('cooldown:arm') && !dimension.getBlock({ x: x + 1, y: y, z: z }).hasTag('adaptable_arm_1') && !dimension.getBlock({ x: x + 2, y: y, z: z }).hasTag('adaptable_arm_2')) itemMovement(3, dimension, x - 1, y, z, block, x, y, z, entity)
                                break
                        }
                    }
                }
            } catch (e) { }
        }
    }
    if (t % 2 === 0) {
        for (let i = 0; i < worldEntities.length; i++) {
            if (worldEntities[i].logCount >= 64) {
                worldEntities[i] = { entityName: worldEntities[i].entityName, idBlock: '', logBlocks: [], otherBlocks: [], logCount: 1, otherCount: 0 }
            }
            else if (worldEntities[i].logBlocks[0] || worldEntities[i].otherBlocks[0]) {
                if (worldEntities[i].logBlocks[0]) {
                    function callsSpeed(Num) {
                        for (let ti = 0; ti < Num; ti++) {
                            treeDestroy(i, worldEntities[i].idBlock, worldEntities[i].logBlocks)
                        }
                    }
                    callsSpeed(5)
                }
                if (worldEntities[i].otherBlocks[0]) {
                    function callsSpeed(Num) {
                        for (let ti = 0; ti < Num; ti++) {
                            treeDestroy(i, worldEntities[i].idBlock, worldEntities[i].otherBlocks)
                        }
                    }
                    callsSpeed(78)
                }
            }
            else worldEntities[i] = { entityName: worldEntities[i].entityName, idBlock: '', logBlocks: [], otherBlocks: [], logCount: 1, otherCount: 0 }
        }
        for (const worlds of dimensions) {
            try {
                for (const entity of world.getDimension(worlds).getEntities({ excludeTypes: ['minecraft:item'] })) {
                    const dimension = entity.dimension
                    let { x, y, z } = entity.location
                    x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                    if (entity.typeId === 'entity:conveyor_belt_filter') {
                        const permutation = dimension.getBlock({ x, y, z }).permutation
                        if (entity.getTags()[0]) {
                            for (const filter of dimension.getEntitiesAtBlockLocation({ x, y, z })) {
                                if (filter.getComponent('minecraft:item')) {
                                    const { x, y, z } = filter.location;
                                    const inverted = permutation.getState('filter:inverted') ? true : false
                                    switch (permutation.getState('block:rotation')) {
                                        case 2: filterItem(x, y, z, entity.getTags(), filter, x, z + 0.8, inverted); break
                                        case 3: filterItem(x, y, z, entity.getTags(), filter, x, z - 0.8, inverted); break
                                        case 4: filterItem(x, y, z, entity.getTags(), filter, x + 0.8, z, inverted); break
                                        case 5: filterItem(x, y, z, entity.getTags(), filter, x - 0.8, z, inverted); break
                                    }
                                }
                            }
                        }
                        function filterItem(x, y, z, itF, f, ix, iz, inverted) {
                            try {
                                const typeId = f.getComponent('minecraft:item').itemStack.typeId
                                if ((itF.every(e => { return !typeId.includes(e.trim()) }) && inverted || itF.some(s => { return typeId === s.trim() }) && !inverted) 
                                && y <= Math.floor(y) + 0.2 &&
                                (x - 0.3 <= Math.floor(x) + 0.5 && x + 0.3 >= Math.floor(x) + 0.5) &&
                                (z - 0.3 <= Math.floor(z) + 0.5 && z + 0.3 >= Math.floor(z) + 0.5)
                                ) {
                                    f.teleport({ x: ix, y: y + 0.26, z: iz }, { dimension: dimension })
                                }
                            } catch (e) { }
                        }
                    }
                }
            } catch (e) { }
        }
    }
    for (const worlds of dimensions) {
        try {
            for (const entity of world.getDimension(worlds).getEntities({ type: 'minecraft:item' })) {
                let { x, y, z } = entity.location
                x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                if (y >= -64 && !entity.getTags().some(e => { return e.includes('arm') })) {
                    try {
                        const block = entity.dimension.getBlock({ x, y, z })
                        if (block.typeId.includes('conveyor_belt')) {
                            let xDiff = (x + 0.5) - entity.location.x
                            let zDiff = (z + 0.5) - entity.location.z
                            let xVelocity = (block.typeId.includes('west') || block.typeId.includes('east')) ? 0 : xDiff / 16
                            let yVelocity = 0
                            let zVelocity = (block.typeId.includes('north') || block.typeId.includes('south')) ? 0 : zDiff / 16
                            switch (block.typeId) {
                                case 'effect99:conveyor_belt_north': zVelocity += 0.09; break
                                case 'effect99:conveyor_belt_south': zVelocity -= 0.09; break
                                case 'effect99:conveyor_belt_west': xVelocity += 0.09; break
                                case 'effect99:conveyor_belt_east': xVelocity -= 0.09; break
                            }
                            if (block.hasTag('sloping_conveyor_top')) yVelocity = -0.035
                            else if (block.hasTag('sloping_conveyor_down')) yVelocity = 0.13
                            const currentV = entity.getVelocity()
                            const desiredV = { x: xVelocity, y: yVelocity, z: zVelocity }
                            const force = { x: desiredV.x - currentV.x, y: desiredV.y - currentV.y, z: desiredV.z - currentV.z }
                            entity.applyImpulse(force);
                        }
                    } catch (e) { }
                }
            }
        } catch (e) { }
    }
    if (t % 18 === 0) {
        try {
            for (const player of world.getAllPlayers()) {
                const { x, y, z } = player.location
                let radiationFound = 0
                for (const r of player.dimension.getEntities({ type: 'entity:radiation_explode', location: { x, y, z }, maxDistance: 48 })) {
                    radiationFound = 1
                }
                if (!radiationFound) player.runCommand('fog @s pop radiation')
            }
        } catch (e) { }
        for (const worlds of dimensions) {
            try {
                for (const entity of world.getDimension(worlds).getEntities({ type: 'entity:storage_core', excludeFamilies: ['StopMachine'] })) {
                    const dimension = entity.dimension
                    let { x, y, z } = entity.location
                    x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                    const core = dimension.getBlock({ x, y, z }).permutation
                    if (entity.typeId === 'entity:storage_core' && core.getState('energy:level')) {
                        const positions = [
                            { panel: { x, z: z - 2, itemsEntry: [{ x, z: z - 2 }, { x: x - 1, z: z - 2 }, { x: x + 1, z: z - 2 }] } },
                            { panel: { x, z: z + 2, itemsEntry: [{ x, z: z + 2 }, { x: x - 1, z: z + 2 }, { x: x + 1, z: z + 2 }] } },
                            { panel: { x: x - 2, z, itemsEntry: [{ x: x - 2, z }, { x: x - 2, z: z - 1 }, { x: x - 2, z: z + 1 }] } },
                            { panel: { x: x + 2, z, itemsEntry: [{ x: x + 2, z }, { x: x + 2, z: z - 1 }, { x: x + 2, z: z + 1 }] } }
                        ]
                        let num = 0
                        let itemSaved = []
                        try {
                            itemSaved = entity.getTags().map(item => JSON.parse(item))
                            num = itemSaved.reduce((add, item) => add + item.amount, 0)
                        } catch (e) { }
                        const math = num && num <= 100000 ? Math.ceil((num - 9999) / 10000) : num >= 99999 ? 10 : 0
                        if (math != core.getState('storage:level')) {
                            dimension.runCommand(setPermutations(x, y, z, core.type.id, core.getAllStates(), { "storage:level": math }))
                        }
                        for (const { panel: { x, z, itemsEntry } } of positions) {
                            const panel = dimension.getBlock({ x: x, y: y - 1, z: z }).permutation
                            let inUse = 0
                            for (const { x, z } of itemsEntry) {
                                if (panel.getState('storage:mode') === 0 && num <= 100000) { /* 90000 */
                                    for (const item of dimension.getEntitiesAtBlockLocation({ x: x, y: y - 2, z: z })) {
                                        if (item.getComponent('minecraft:item')) {

                                            inUse = 1
                                            const itemStack = item.getComponent('minecraft:item').itemStack
                                            const { typeId, amount } = itemStack
                                            const lore = (typeId === 'effect99:item_hoover' && itemStack.getLore()[0]) ? Number(itemStack.getLore()[0].match(/[\d.]+/g)[0]) : 0
                                            const dur = itemStack.getComponent("durability") ? itemStack.getComponent("durability").damage : lore ? lore : 0
                                            const enchantments = itemStack.getComponent("enchantable") ? itemStack.getComponent("enchantable").getEnchantments() : []
                                            let enchant = []

                                            for (const enchantment of enchantments) {
                                                enchant.push({ enchant: enchantment.type.id, level: enchantment.level })
                                            }

                                            let matched = false
                                            itemSaved = entity.getTags().map(item => JSON.parse(item))
                                            for (const i of itemSaved) {
                                                if (i.typeId === typeId && i.dur === dur && matchedEnchant(i.ench, enchant)) {
                                                    matched = true
                                                    entity.removeTag(JSON.stringify(i))
                                                    entity.addTag(JSON.stringify({ typeId, amount: i.amount + amount, dur, ench: enchant }))
                                                    break
                                                }
                                            }
                                            if (!matched) {
                                                entity.addTag(JSON.stringify({ typeId, amount, dur: lore ? lore : dur, ench: enchant }))
                                            }
                                            item.kill()

                                        }
                                    }
                                }
                                else if (panel.getState('storage:mode')) {
                                    try {
                                        if (entity.getTags()[0]) {
                                            inUse = 1
                                            const rNum = Math.floor(Math.random() * entity.getTags().length)
                                            let itemSaved = entity.getTags()[0] ? JSON.parse(entity.getTags()[rNum]) : undefined
                                            if (itemSaved) {
                                                const { typeId, amount, dur, ench } = itemSaved
                                                let maxStack = entity.getComponent('minecraft:mark_variant').value
                                                const lore = typeId === 'effect99:item_hoover' ? dur : 0
                                                const itemStack = new ItemStack(typeId, maxStack)
                                                maxStack = Math.min(amount, maxStack, itemStack.maxAmount)
                                                itemStack.amount = maxStack

                                                if (amount === 1 || amount - maxStack === 0) {
                                                    entity.removeTag(JSON.stringify(itemSaved))
                                                } else {
                                                    entity.removeTag(JSON.stringify(itemSaved))
                                                    entity.addTag(JSON.stringify({ typeId, amount: amount - maxStack, dur, ench }))
                                                }

                                                dur && !lore ? itemStack.getComponent("durability").damage = dur : 0
                                                const enchantments = itemStack.getComponent("enchantable")
                                                for (const { enchant, level } of ench) {
                                                    enchantments.addEnchantment({ type: enchant, level: level })
                                                }
                                                if (lore) itemStack.setLore([`§r§lBattery ${dur}‰`, '§r§l§cOff'])

                                                const item = dimension.spawnItem(itemStack, { x: x + 0.5, y: y - 2, z: z + 0.5 })
                                                item.teleport({ x: x + 0.5, y: y - 1.9, z: z + 0.5 }, { dimension: dimension, keepVelocity: false })
                                                switch (panel.getState('block:rotation')) {
                                                    case 0: item.applyImpulse({ x: 0, y: 0.05, z: 0.15 }); break
                                                    case 1: item.applyImpulse({ x: 0, y: 0.05, z: -0.15 }); break
                                                    case 2: item.applyImpulse({ x: 0.15, y: 0.05, z: 0 }); break
                                                    case 3: item.applyImpulse({ x: -0.15, y: 0.05, z: 0 }); break
                                                }
                                            }
                                        }
                                    } catch (e) { }
                                }
                                
                            }
                            try {
                                if (inUse) {
                                    if (entity.getComponent('minecraft:skin_id').value < 100) {
                                        entity.getComponent('minecraft:skin_id').value++
                                    }
                                    else {
                                        let { x, y, z } = entity.location
                                        x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                                        entity.getComponent('minecraft:skin_id').value = 0
                                        const permutation = entity.dimension.getBlock({ x, y, z }).permutation
                                        entity.dimension.runCommand(setPermutations(x, y, z, permutation.type.id, permutation.getAllStates(), { "energy:level": permutation.getState('energy:level') - 1 }))
                                    }
                                }
                            } catch (e) { }
                        }
                    }
                }
            } catch (e) { }
        }
    }
    if (itemMovementTicks) {
        itemMovementTicks -= 1
        itemsInMovement.forEach(({ item, location, dimension, t }, i) => {
            try {
                let { x, y, z } = location
                x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
                if (t >= 9) {
                    item.teleport({ x: x + 0.5, y: y + 0.7, z: z + 0.5 }, { dimension: dimension, keepVelocity: false })
                }
                if (item.hasTag('arm0') && t) {
                    item.applyImpulse({ x: 0, y: 0.038, z: -0.024 })
                    itemsInMovement[i].t -= 1
                }
                else if (item.hasTag('arm1') && t) {
                    item.applyImpulse({ x: 0, y: 0.038, z: 0.024 })
                    itemsInMovement[i].t -= 1
                }
                else if (item.hasTag('arm2') && t) {
                    item.applyImpulse({ x: -0.024, y: 0.038, z: 0 })
                    itemsInMovement[i].t -= 1
                }
                else if (item.hasTag('arm3') && t) {
                    item.applyImpulse({ x: 0.024, y: 0.038, z: 0 })
                    itemsInMovement[i].t -= 1
                }
                else {
                    for (const tag of item.getTags()) {
                        item.removeTag(tag)
                    }
                    itemsInMovement.splice(i, 1)
                }
            } catch (e) { }
        })
    }
    worldTicks()
})
worldTicks()

function matchedEnchant(a, b) {
    if (a.length !== b.length) {
        return false;
    }
    return a.every(item => b.some(
        (otherItem) => JSON.stringify(item) === JSON.stringify(otherItem)
    ))
}

function itemMovement(r, dimension, x, y, z, block, bx, by, bz, entity) {
    for (const item of dimension.getEntitiesAtBlockLocation({ x, y, z })) {
        const { x: ix, y: iy, z: iz } = item.location
        if (item.getComponent('minecraft:item') && iy <= Math.floor(iy) + 0.2 && (ix - 0.3 <= Math.floor(ix) + 0.5 && ix + 0.3 >= Math.floor(ix) + 0.5) && (iz - 0.3 <= Math.floor(iz) + 0.5 && iz + 0.3 >= Math.floor(iz) + 0.5)) {
            switch (r) {
                case 0: item.addTag('arm0'); break
                case 1: item.addTag('arm1'); break
                case 2: item.addTag('arm2'); break
                case 3: item.addTag('arm3'); break
            }
            itemMovementTicks = 20
            dimension.runCommand(setPermutations(bx, by, bz, block.type.id, block.getAllStates(), { "cooldown:arm": 1 }))
            entity.runCommand(`playanimation @s animation.${entity.typeId.slice(entity.typeId.indexOf(":") + 1)}.move`)
            itemsInMovement.push({ item: item, location: item.location, dimension, t: 10 })
        }
    }
}
function fuelProvider(block, x, y, z, e, ex, ey, ez, dimension, bx, by, bz) {
    let i = 0
    for (const item of e.dimension.getEntitiesAtBlockLocation({ x: ex, y: ey, z: ez })) {
        if (i) break; i++
        const { typeId, amount } = item.getComponent('minecraft:item').itemStack
        if (fuelTypeA.includes(typeId) || fuelTypeB.includes(typeId) || fuelTypeC.includes(typeId) || fuelTypeD.includes(typeId) || (typeId.includes('minecraft:') && keywordFuel.some(i => { return typeId.includes(i) }))) {
            const { x: ix, y: iy, z: iz } = item.location
            const newStack = item.getComponent('minecraft:item').itemStack.clone()
            newStack.amount > 1 ? newStack.amount -= 1 : 0
            if (fuelTypeA.includes(typeId)) {
                e.runCommand(`playanimation @s animation.${e.typeId.slice(e.typeId.indexOf(":") + 1)}.move`); dimension.spawnEntity('entity:fuel_provider<fuel_1>', { x: bx, y: by, z: bz })
                dimension.runCommand(setPermutations(x, y, z, block.type.id, block.getAllStates(), { "cooldown:arm": 1 }))
                if (amount - 1) {
                    dimension.spawnItem(newStack, { x: ix, y: iy, z: iz })
                    item.kill()
                } else item.kill()
            }
            else if (fuelTypeB.includes(typeId)) {
                e.runCommand(`playanimation @s animation.${e.typeId.slice(e.typeId.indexOf(":") + 1)}.move`); dimension.spawnEntity('entity:fuel_provider<fuel_2>', { x: bx, y: by, z: bz })
                dimension.runCommand(setPermutations(x, y, z, block.type.id, block.getAllStates(), { "cooldown:arm": 1 }))
                if (amount - 1) {
                    dimension.spawnItem(newStack, { x: ix, y: iy, z: iz })
                    item.kill()
                } else item.kill()
            }
            else if (fuelTypeC.includes(typeId)) {
                e.runCommand(`playanimation @s animation.${e.typeId.slice(e.typeId.indexOf(":") + 1)}.move`); dimension.spawnEntity('entity:fuel_provider<fuel_3>', { x: bx, y: by, z: bz })
                dimension.runCommand(setPermutations(x, y, z, block.type.id, block.getAllStates(), { "cooldown:arm": 1 }))
                if (amount - 1) {
                    dimension.spawnItem(newStack, { x: ix, y: iy, z: iz })
                    item.kill()
                } else item.kill()
            }
            else if (fuelTypeD.includes(typeId) || (typeId.includes('minecraft:') && keywordFuel.some(i => { return typeId.includes(i) }))) {
                e.runCommand(`playanimation @s animation.${e.typeId.slice(e.typeId.indexOf(":") + 1)}.move`); dimension.spawnEntity('entity:fuel_provider<fuel_4>', { x: bx, y: by, z: bz })
                dimension.runCommand(setPermutations(x, y, z, block.type.id, block.getAllStates(), { "cooldown:arm": 1 }))
                if (amount - 1) {
                    dimension.spawnItem(newStack, { x: ix, y: iy, z: iz })
                    item.kill()
                } else item.kill()
            }
        }
    }
}

world.afterEvents.entitySpawn.subscribe(({ entity }) => {
    try {
        if (entity.typeId === 'entity:treecapitator') {
            worldEntities.push({ entityName: entity.id, idBlock: '', logBlocks: [], otherBlocks: [], logCount: 1, otherCount: 0 })
            let { x, y, z } = entity.location
            x = Math.floor(x); y = Math.floor(y); z = Math.floor(z)
            const blockId = entity.dimension.getBlock({ x: x, y: y + 1, z: z }).typeId
            if ((blockId.includes('log') && !blockId.includes('stripped'))) {
                for (let i = 0; i < worldEntities.length; i++) {
                    if (worldEntities[i].entityName == entity.id) {
                        worldEntities[i].logBlocks.push(`${x} ${y + 1} ${z}`)
                        worldEntities[i].idBlock = blockId
                    }
                }
            }
        }
    } catch (e) { }
})

function treecapitator(block, x, y, z, e) {
    try {
        const blockSides = [[x, y - 1, z], [x + 1, y, z], [x, y, z - 1], [x - 1, y, z], [x, y, z + 1], [x, y + 1, z], [x + 1, y + 1, z + 1], [x + 1, y + 1, z - 1], [x - 1, y + 1, z - 1], [x - 1, y + 1, z + 1]]
        for (const coords of blockSides) {
            const idBlocks = world.getDimension("overworld").getBlock({ x: coords[0], y: coords[1], z: coords[2] }).typeId
            if (idBlocks == block && worldEntities[e].logCount < 64) {
                if (worldEntities[e].logBlocks.every(l => l != `${coords[0]} ${coords[1]} ${coords[2]}`)) {
                    worldEntities[e].logBlocks.push(`${coords[0]} ${coords[1]} ${coords[2]}`)
                    worldEntities[e].logCount += 1
                }
            }
            if ((idBlocks.includes('leaves') || idBlocks.includes('vine')) && (worldEntities[e].otherCount < worldEntities[e].logCount * 6 && worldEntities[e].otherCount < 290)) {
                if (worldEntities[e].otherBlocks.every(o => o != `${coords[0]} ${coords[1]} ${coords[2]}`)) {
                    worldEntities[e].otherBlocks.push(`${coords[0]} ${coords[1]} ${coords[2]}`)
                    worldEntities[e].otherCount += 1
                }
            }
        }
    } catch (e) { }
}
function treeDestroy(e, b, list) {
    try {
        if (worldEntities[e].logBlocks[0] || worldEntities[e].otherBlocks[0]) {
            const x = Number(list[0].slice(0, list[0].indexOf(' ')))
            const y = Number(list[0].slice(list[0].indexOf(' '), list[0].indexOf(' ', list[0].indexOf(' ') + 1)))
            const z = Number(list[0].slice(list[0].indexOf(' ', list[0].indexOf(' ') + 1)))
            //world.getDimension("overworld").runCommand(`title @a actionbar ${b} ${worldEntities[e].logCount} : ${worldEntities[e].otherCount} | ${x} ${y} ${z} | \nIndex Entities: ${e}`)
            treecapitator(b, x, y, z, e)
        }
        world.getDimension("overworld").runCommand(`setblock ${list.splice(0, 1)} air 0 destroy`)
    } catch (e) { }
}

function give(player, item, message = false) {
    player.runCommand(`replaceitem entity @s slot.weapon.mainhand 0 ${item}`)
    if (message) player.onScreenDisplay.setActionBar(`You do not have permission\n to have this item`)
}

world.beforeEvents.chatSend.subscribe(chat => {
    if (chat.message === '!admin') {
        chat.cancel = true
        admin = true
        world.sendMessage(`Admin ${chat.sender.name}`)
    } else if (chat.message === '!admin!') {
        chat.cancel = true
        admin = false
        world.sendMessage(`User ${chat.sender.name}`)
    } else if (chat.message === '!addonV') {
        chat.cancel = true
        world.sendMessage(`§6Add-On §fVersion Beta 0.3.9.3 - §6Minecraft §f1.20.60`)
    }
})

world.afterEvents.worldInitialize.subscribe(() => {
    system.runTimeout(() => {
        for (const dimension of dimensions) {
            for (const item of world.getDimension(dimension).getEntities({ type: 'minecraft:item' })) {
                ['arm0', 'arm0', 'arm0', 'arm0'].every(i => {
                    if (item.hasTag(i)) {
                        item.removeTag(i)
                    }
                })
            }
        }
    }, 30)
})

const fuelTypeA = [
    'minecraft:coal_block',
    'minecraft:blaze_rod',
    'minecraft:lava_bucket'
]

const fuelTypeB = [
    'minecraft:coal',
    'minecraft:charcoal',
    'minecraft:dried_kelp_block',
    'minecraft:blaze_powder'
]

const fuelTypeC = [
    'minecraft:planks',
    'minecraft:log',
    'minecraft:log2',
    'minecraft:planks',
    'minecraft:wood',
    'miencraft:stripped_cherry_log',
    'minecraft:stripped_acacia_log',
    'minecraft:stripped_birch_log',
    'minecraft:stripped_dark_oak_log',
    'minecraft:stripped_jungle_log',
    'minecraft:stripped_oak_log',
    'minecraft:stripped_spruce_log',
    'minecraft:mangrove_wood',
    'minecraft:wooden_door',
    'minecraft:wool',
    'minecraft:lectern',
    'minecraft:chest',
    'minecraft:trapped_chest',
    'minecraft:barrel',
    'minecraft:composter',
    'minecraft:crafting_table',
    'minecraft:bookshelf',
    'minecraft:noteblock'
]

const fuelTypeD = [
    'minecraft:stick',
    'minecraft:ladder',
    'minecraft:deadbush',
    'minecraft:sapling'
]

const keywordFuel = [
    'cherry',
    'jungle',
    'acacia',
    'birch',
    'dark_oak',
    'spruce',
    'mangrove',
    'oak',
    'wooden',
    'leaves',
    'azalea'
]